/**
 * \file stars.c
 * \brief Game of Life Demo
 *
 * lionel_debroux@yahoo.fr (TICT) 2010-2011
 *
 * Port to the Nspire (Ndless pre-2.0) of the TI-68k version made in 2000
 * by Thomas Nussbaumer.
 *
 * This program is free software. It comes without any warranty, to
 * the extent permitted by applicable law. You can redistribute it
 * and/or modify it under the terms of the Do What The Fuck You Want
 * To Public License, Version 2, as published by Sam Hocevar. See
 * http://sam.zoy.org/wtfpl/COPYING for more details.
 *
 *-----------------------------------------------------------------------------
 * The Game of Life was invented by John Conway (as you might have gathered). 
 * The game is played on a field of cells, each of which has eight neighbors 
 * (adjacent cells). A cell is either occupied (by an organism) or not. 
 *
 * The rules for deriving a generation from the previous one are these: 
 * ====================================================================
 *
 * [Death]    If an occupied cell has 0, 1, 4, 5, 6, 7, or 8 occupied neighbors, 
 *            the organism dies
 *            (0, 1 neighbors: of loneliness; 4 thru 8: of overcrowding). 
 *
 * [Survival] If an occupied cell has two or three neighbors, the organism 
 *            survives to the next generation.
 *
 * [Birth]    If an unoccupied cell has three occupied neighbors, it becomes 
 *            occupied.  
 *-----------------------------------------------------------------------------
 *
 */

//-----------------------------------------------------------------------------
// Implementation notes
//-----------------------------------------------------------------------------
// * To keep things simple this program allocates two buffers which hold one
//   pixel per byte. This way we don't have to perform repeated bit extraction.
//   The penalty for this is, of course, an increased memory usage.
//   Other than this, the program is not particularly optimized for speed.
//   It's just a 30 minutes "what-can-I-do-with-my-lunch-break" hack.
//
// * The handling of the 4 corner points is left out for simplicity !!!!
//
//                                                             - Tom.
//-----------------------------------------------------------------------------

#include <os.h>
#include <libndls.h>
#include <inttypes.h>


#define WAIT(n, name) \
  asm volatile ( \
    "ldr r0, ="n"\n" \
    "_loop_"name":\n" \
    "subs r0, r0, #1\n" \
    "bne _loop_"name"\n" \
    : : : "r0" );


uint8_t * source = 0;
uint8_t * dest   = 0;
#define maxcol   (SCREEN_WIDTH - 1)
#define maxline  (SCREEN_HEIGHT - 1)
//uint16_t maxcol  = 0;
//uint16_t maxline = 0;

#define SVAL(l,c)  *(source+SCREEN_WIDTH*(l)+(c))
#define DVAL(l,c)  *(dest+SCREEN_WIDTH*(l)+(c))


//! Do one iteration
void Process (void) {
    //unsigned int i;
    //unsigned char* in  = source;
    //unsigned char* out = dest;

    uint32_t line, col;

    //-------------------------------------------------
    // handle inner content (border will handled later)
    //-------------------------------------------------
    for (line = 1; line < maxline; line++) {
        for (col = 1; col < maxcol; col++) {
            uint32_t sum = SVAL(line-1,col-1) + SVAL(line-1,col) + SVAL(line-1,col+1);
            sum += SVAL(line,col-1) + SVAL(line,col+1);
            sum += SVAL(line+1,col-1) + SVAL(line+1,col) + SVAL(line+1,col+1);

            if (SVAL(line,col)) {
                //-------------------
                // die or survive ???
                //-------------------
                if (sum > 1 && sum < 4) DVAL(line,col) = 1;
                else                    DVAL(line,col) = 0;
            }
            else {
                //-------------------------------------------
                // birth of a new member of this universe ??
                // or still a waste cell ??
                //-------------------------------------------
                if (sum == 3) DVAL(line,col) = 1;
                else          DVAL(line,col) = 0;
            }
        }
    }

    //-------------------------------------------------
    // handle top line with wrap around to bottom line
    // AND
    // handle bottom line with wrap around to top line
    //-------------------------------------------------
    for (col = 1; col < maxcol; col++) {
        uint32_t sumtop, sumbottom;

        sumtop = SVAL(maxline,col-1) + SVAL(maxline,col) + SVAL(maxline,col+1) +
                 SVAL(0,col-1)       + SVAL(0,col+1)     +
                 SVAL(1,col-1)       + SVAL(1,col)       + SVAL(1,col+1);

        sumbottom = SVAL(maxline-1,col-1) + SVAL(maxline-1,col) + SVAL(maxline-1,col+1) +
                    SVAL(maxline,col-1)   + SVAL(maxline,col+1) +
                    SVAL(0,col-1)         + SVAL(0,col)         + SVAL(0,col+1);

        if (SVAL(0,col)) {
            if (sumtop > 1 && sumtop < 4) DVAL(0,col) = 1;
            else                          DVAL(0,col) = 0;
        }
        else {
            if (sumtop == 3) DVAL(0,col) = 1;
            else             DVAL(0,col) = 0;
        }

        if (SVAL(maxline,col)) {
            if (sumbottom > 1 && sumbottom < 4) DVAL(maxline,col) = 1;
            else                                DVAL(maxline,col) = 0;
        }
        else {
            if (sumbottom == 3) DVAL(maxline,col) = 1;
            else                DVAL(maxline,col) = 0;
        }
    }

    //--------------------------------------------------------------
    // handle most-left column with wrap around to most-right column
    // AND
    // handle most-right column with wrap around to most-left column
    //--------------------------------------------------------------
    for (line = 1; line < maxline; line++) {
        uint32_t sumleft,sumright;

        sumleft  = SVAL(line-1,maxcol) + SVAL(line-1,0) + SVAL(line-1,1) +
                   SVAL(line,  maxcol) + SVAL(line,1) +
                   SVAL(line+1,maxcol) + SVAL(line+1,0) + SVAL(line+1,1);

        sumright = SVAL(line-1,maxcol-1) + SVAL(line-1,maxcol) + SVAL(line-1,0) +
                   SVAL(line,  maxcol-1) + SVAL(line,0) +
                   SVAL(line+1,maxcol-1) + SVAL(line+1,maxcol) + SVAL(line+1,0);

        if (SVAL(line,0)) {
            if (sumleft > 1 && sumleft < 4) DVAL(line,0) = 1;
            else                            DVAL(line,0) = 0;
        }
        else {
            if (sumleft == 3) DVAL(line,0) = 1;
            else              DVAL(line,0) = 0;
        }

        if (SVAL(line,maxcol)) {
            if (sumright > 1 && sumright < 4) DVAL(line,maxcol) = 1;
            else                              DVAL(line,maxcol) = 0;
        }
        else {
            if (sumright == 3) DVAL(line,maxcol) = 1;
            else               DVAL(line,maxcol) = 0;
        }
    }
}


//! Build our special "one pixel per byte" array from original LCD content
void GrabScreen(void) {
    uint8_t * lcdmem = SCREEN_BASE_ADDRESS;    
    uint8_t * tmp    = source;
    uint32_t  i;

    // This code assumes 4 bits per pixel, white <-> 1111 and black <-> 0000.
    for (i = 0; i < SCREEN_BYTES_SIZE; i++) {
        if (!(*lcdmem & 0x0F)) *tmp++ = 1;
        else                   *tmp++ = 0;
        if (!(*lcdmem & 0xF0)) *tmp++ = 1;
        else                   *tmp++ = 0;
        lcdmem++;
    }
}    


//! Write our special "one pixel per byte" array to LCD buffer
void PushScreen(void) {
    uint8_t * lcdmem = SCREEN_BASE_ADDRESS;    
    uint8_t * tmp    = dest;
    uint32_t  i;
    
    // This code assumes 4 bits per pixel, white <-> 1111 and black <-> 0000.
    for (i = 0; i < SCREEN_BYTES_SIZE; i++) {
        uint8_t val = 0xFF;
        if (*tmp++) val &= 0x0F;
        if (*tmp++) val &= 0xF0;
        *lcdmem++ = val;
    }
}    


//! Where all the fun starts, as usual...
int main(void) {
    uint32_t counter = 0;
    char  outbuf[100];
    uint8_t * memblock;

    show_msgbox(
        "life - Game of Life demo",
        "------------------------\n"
        "    Game of Life demo\n"
        "    © 2000-2010 TICT\n"
        "------------------------\n"
        "ESC   - Exit"
    );

    WAIT("0x1FFFFF", "wait1");

    //------------------------------------------------------
    // allocate two "one pixel per byte" array buffers
    //------------------------------------------------------
    if ((memblock = malloc(SCREEN_WIDTH * SCREEN_HEIGHT * 2)) != NULL) {

        source = memblock;
        dest = source + SCREEN_WIDTH * SCREEN_HEIGHT;

        //maxcol  = SCREEN_WIDTH  - 1;
        //maxline = SCREEN_HEIGHT - 1;

        //------------------------------------------------------
        // fill up source array with data from screen
        //------------------------------------------------------
        GrabScreen();

        //------------------------------------------------------
        // loop until user gets bored ...
        //------------------------------------------------------
        while (!isKeyPressed(KEY_NSPIRE_ESC)) {
           uint8_t * tmp;

           //-----------------------------------------------
           // do a full iteration, push data back on screen,
           // output the iteration counter and swap source
           // and dest pointers
           //-----------------------------------------------
           Process();
           PushScreen();
           counter++;
           //sprintf(outbuf,"%05u",counter);
           //DrawStrXY(0,0,outbuf,A_REPLACE);
           tmp = dest;
           dest = source;
           source = tmp;
        }

        // Free allocated memory and get out here !!
        free(memblock);
    }

    WAIT("0x1FFFFF", "wait2");

    sprintf(outbuf, "Run %" PRIu32 " iterations.", counter);
    show_msgbox(
        "Simulation complete",
        outbuf);

    return 0;
}
